;===============================================================================
;  simple calculator (add/sub/mul/div/sqrt, percent key, memory add/sub)
;  with exchangable input/output interface (e.g. Keypad + LCD/LED/VFD or RS232)
;===============================================================================
;
;Controller: ATTiny2313 @ 8 MHz (internal R/C-oscillator)
;Low  Fuse Byte: 0xC4 -> 8 MHz internal R/C osc., BOD enabled
;High Fuse Byte: 0xDB -> BOD Level = 2.7 V
;
;Ports: see I/O module
;
;Version History:
;  0.1.0  2007-06-16  Arne Rossius
;    * version for testing math functions, RS232 I/O interface only
;  0.3.0  2007-06-18  Arne Rossius
;    * working version with LED display and RS232 input, no mul/div and memory
;  0.4.0  2007-06-29  Arne Rossius
;    * added multiplication, still some rounding errors
;  0.5.0  2007-06-24  Arne Rossius
;    * new multiplication routine with better rounding
;    * remove (-) sign if result is zero
;    * percent: add, sub, copy (missing: mul, div)
;  0.6.0  2007-07-02  Arne Rossius
;    * percent: added mul, div (depending on not yet implemented division)
;    * changed input to 4x6 matrix keypad
;  0.7.0  2007-07-08  Arne Rossius
;    * BUGFIX: stack leak in multiplication routine corrupts return address
;    * BUGFIX: trailing zeros after decimal point after rounding
;    * BUGFIX: minus sign not removed if result is only zero after rounding
;    * added division routine
;    * display sign directly in front of the number if not all digits are used
;    * merged some redundant code into new functions
;  0.8.0  2007-07-09  Arne Rossius
;    * some minor bugfixes
;    * added memory
;    * cleaned up several routines, added second flags register
;    * added new public function in I/O module: 'io_updatesymbols'
;  0.8.1  2007-09-13  Arne Rossius
;    * BUGFIX: lost sign of result when using "CE"
;  0.8.2  2007-12-02  Arne Rossius
;    * updated RS232 I/O-module to include all available functions
;  0.9  2007-12-02  Arne Rossius
;    * BUGFIX: rounding for display failed if a new digit was introduced (e.g. 99.9->100)
;    * changed all numbers to packed BCD to save RAM (fixes bug with DISPLEN=16)
;    * new keypad scanning routine to save Flash memory (needed for sqrt and 1/x)
;    * new keystroke comparision routine (for RS232) to save Flash memory
;    * new percentage add/sub routine to save Flash memory
;    * added two new number registers (num_start, num_iter) for sqrt iteration
;    * square root and reciprocal value functions
;    * additional (7th) row of keys in 'io_led_keypad.asm' for new functions
;    * changed some function names to represent the used number (mostly num_io)
;    * several small bugfixes
;  0.9.1  2007-12-20  Arne Rossius
;    * added alternate keyboard layout in "io_led_keypad.asm"
;  0.9.2  2008-11-26  Arne Rossius
;    * new I/O-module "io_vfd_keypad.asm" (some months ago)
;    * new I/O-module "io_vfd_ps2.asm" for PS/2-keypad
;    * accept '=' immediately after operation (e.g. for x^2, use x * =)
;    * support for constant value calculation (x + a = b = c = ...)
;    * inverse (1/x) function doesn't affect first operand anymore (1 + 5 1/x = 1.2)
;    * TODO: square root clears first operand (no register left to keep it safe!)
;  0.9.3  2008-12-23  Arne Rossius
;    * merged I/O modules "io_led_keypad" and "io_vfd_keypad" into "io_7seg_keypad"

.include "../include/tn2313def.inc"

;R0-R15 = reserved for I/O module
.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	flags = R19
	.equ	num_io_neg = 0 ;I/O data is negative
	.equ	num_op1_neg = 1 ;operand 1 is negative
	.equ	num_temp_neg = 2 ;temp number is negative
	.equ	ManualInput = 3 ;user input is present
	.equ	AnyInput = 4 ;any input is present (user or memory recall)
	.equ	InputNonzero = 5 ;entered any of: 1 2 3 4 5 6 7 8 9 dot
	.equ	Error = 6 ;calculation returned an error
	.equ	ShiftRounded = 7 ;last right-shift rounded *up*
.def	flags2 = R20
	.equ	MemNeg = 0 ;number in memory is negative
	.equ	MemZero = 1 ;number in memory is zero
	.equ	num_start_neg = 2 ;iteration start number is negative
	.equ	OP_EQUAL = 3 ;last operator key was "equal"
	.equ	OP_ADD = 4 ;next operation will be addition
	.equ	OP_SUB = 5 ;next operation will be subtraction
	.equ	OP_MUL = 6 ;next operation will be multiplication
	.equ	OP_DIV = 7 ;next operation will be division
.def	inplen = R21
.def	decimals = R22
.def	first = R23
.def	last = R24
.def	length = R25
;R27:R26 = X pointer (reserved for I/O module)
;R29:R28 = Y pointer
;R31:R30 = Z pointer

;Keys '0' to '9' = 0x00 to 0x09
.equ	KEY_ADD = 0x10
.equ	KEY_SUB = 0x11
.equ	KEY_MUL = 0x12
.equ	KEY_DIV = 0x13
.equ	KEY_PERCENT = 0x14
.equ	KEY_EQUAL = 0x15
.equ	KEY_NEG = 0x16
.equ	KEY_CE = 0x17
.equ	KEY_AC = 0x18
.equ	KEY_DOT = 0x19
;optional memory keys:
.equ	KEY_MEMADD = 0x20
.equ	KEY_MEMSUB = 0x21
.equ	KEY_MEMRCL = 0x22
.equ	KEY_MEMCLR = 0x23
;optional special function keys:
.equ	KEY_INV = 0x40
.equ	KEY_SQRT = 0x41

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
;.include "io_rs232.asm"
;.include "io_led_keypad.asm" ;OBSOLETE
;.include "io_vfd_keypad.asm" ;OBSOLETE
.include "io_7seg_keypad.asm"
;.include "io_vfd_ps2.asm"
.include "math.asm"

;===============================================================================

.dseg
	num_io: .byte DISPSIZE
	num_op1: .byte DISPSIZE
	num_temp: .byte DISPSIZE
	num_start: .byte DISPSIZE
	num_iter: .byte DISPSIZE
	mem: .byte DISPSIZE
	stack: .byte 16
	
;===============================================================================

.cseg
reset:
	;init stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;initialize I/O devices
	rcall	io_init
	
	;load 0 to memory
	rcall	clear_mem
	
init:
	;initialize registers and load 0 to num_op1/num_io/num_temp
	rcall	clear_input
	ldi	flags, 1<<ManualInput ;prevent (-) getting lost if pressed first
	andi	flags2, 1<<MemZero | 1<<MemNeg ;memory wasn't changed
	rcall	clear_op1
	rcall	clear_io
	
;-------------------------------------------------------------------------------

update: ;round num_io to fit display size
	rcall	copy_io_to_temp ;backup original precision
update_loop:
	;get position of first displayed (nonzero) digit
	ldi	ZL, num_io
	rcall	numfirst
	brne	PC+2
	andi	flags, ~(1<<num_io_neg) ;no (-) sign if result is zero
	cpi	first, DISPSIZE ;display at least 1 digit before decimal point
	brlo	PC+2
	ldi	first, DISPSIZE-1
	;get position of last displayed (nonzero) digit after decimal point
	ldi	last, DISPSIZE*2-1
	ldi	ZL, num_io+DISPSIZE
update_last_loop:
	ld	temp, -Z
	swap	temp
	cpi	temp, 0x10
	brsh	update_last_end
	dec	last
	cpi	temp, 0
	brne	update_last_end
	dec	last
	cpi	last, DISPSIZE ;display at least up to units digit
	brsh	update_last_loop
update_last_end:
	;calculate length
	mov	length, last
	sub	length, first
	inc	length
	
	;round if length > DISPSIZE
	andi	flags, ~(1<<ShiftRounded)
	cpi	length, DISPSIZE+1
	brlo	update_round_end
	;round
	ldi	temp, DISPSIZE
	add	temp, first ;temp = first digit that isn't displayed
	mov	temp2, temp
	lsr	temp2
	ldi	ZL, num_io
	add	ZL, temp2
	ld	temp2, Z
	sbrc	temp, 0
	swap	temp2
	cpi	temp2, 0x50
	brlo	update_round_end
	;round up: add 5 to first digit that isn't displayed
	push	ZL
	rcall	copy_op1_to_start
	rcall	clear_op1
	pop	ZL
	ldi	temp2, 0x50
	sbrc	temp, 0
	ldi	temp2, 0x05
	std	Z+(num_op1-num_io), temp2
	rcall	math_add_abs
	rcall	copy_start_to_op1
	rjmp	update_loop ;re-calculate first/last displayed digits
update_round_end:

	;remove trailing zeros from *displayed* fractional part of the result
	cpi	length, DISPSIZE+1 ;limit length to DISPSIZE digits
	brlo	PC+2
	ldi	length, DISPSIZE
	mov	last, first
	add	last, length
	dec	last
	mov	temp2, last
	lsr	temp2
	ldi	ZL, num_io
	add	ZL, temp2
update_trailingzero:
	cpi	last, DISPSIZE
	brlo	update_trailingzero_end
	ld	temp, Z
	sbrc	last, 0
	swap	temp
	cpi	temp, 0x10
	brsh	update_trailingzero_end
	sbrs	last, 0
	dec	ZL
	dec	last
	dec	length
	rjmp	update_trailingzero
update_trailingzero_end:
	;remove (-) sign if *displayed* part of the result is zero
	cpi	length, 1
	brne	update_nonzero
	ldi	ZL, num_io+(DISPSIZE/2-1)
	ld	temp, Z
	cpi	temp, 0x00 ;upper nibble always zero if length=1
	brne	update_nonzero
	andi	flags, ~(1<<num_io_neg)
update_nonzero:
	;update display
	rcall	io_update
	rcall	copy_temp_to_io ;restore original precision

;-------------------------------------------------------------------------------

waitkey:
	rcall	io_getkey
	cpi	temp, KEY_AC
	brne	PC+2
	rjmp	init
	sbrc	flags, Error ;lock all keys but 'AC' if an error occured
	rjmp	waitkey
	cpi	temp, KEY_EQUAL
	brne	PC+2
	rjmp	equal
	cpi	temp, KEY_PERCENT
	brne	PC+2
	rjmp	percent
	cpi	temp, KEY_MEMSUB
	brne	PC+2
	rjmp	memsub
	cpi	temp, KEY_DOT
	breq	dot
	cpi	temp, KEY_NEG
	breq	flip_sign
	cpi	temp, KEY_ADD
	breq	addition
	cpi	temp, KEY_SUB
	breq	subtraction
	cpi	temp, KEY_MUL
	breq	multiplication
	cpi	temp, KEY_DIV
	breq	division
	cpi	temp, KEY_INV
	breq	invert
	cpi	temp, KEY_SQRT
	breq	sqrt
	cpi	temp, KEY_CE
	breq	ce
	cpi	temp, KEY_MEMCLR
	breq	memclr
	cpi	temp, KEY_MEMRCL
	breq	memrcl
	cpi	temp, KEY_MEMADD
	breq	memadd
	cpi	temp, 10
	brsh	waitkey
	rjmp	digit
	
dot:
	;decimal point
	cpi	decimals, 0
	brne	waitkey ;number already has a decimal point
	ldi	decimals, 1
	ori	flags, 1<<AnyInput | 1<<ManualInput | 1<<InputNonzero
	cpi	inplen, 0
	brne	waitkey
	rcall	clear_io
	ldi	inplen, 1
	rjmp	digit_end
	
flip_sign:
	;flip sign of num_io
	ldi	temp, 1<<num_io_neg
	eor	flags, temp
	rjmp	update

addition:
	;perform last requested operation, next operation will be addition
	sbrc	flags2, OP_EQUAL
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
	rcall	perform_operation
	rcall	copy_io_to_op1
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL) ;clear prev operation
	ori	flags2, 1<<OP_ADD
	rjmp	update
	
subtraction:
	;perform last requested operation, next operation will be subtraction
	sbrc	flags2, OP_EQUAL
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
	rcall	perform_operation
	rcall	copy_io_to_op1
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL) ;clear prev operation
	ori	flags2, 1<<OP_SUB
	rjmp	update
	
multiplication:
	;perform last requested operation, next operation will be multiplication
	sbrc	flags2, OP_EQUAL
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
	rcall	perform_operation
	rcall	copy_io_to_op1
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL) ;clear prev operation
	ori	flags2, 1<<OP_MUL
	rjmp	update
	
division:
	;perform last requested operation, next operation will be division
	sbrc	flags2, OP_EQUAL
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
	rcall	perform_operation
	rcall	copy_io_to_op1
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL) ;clear prev operation
	ori	flags2, 1<<OP_DIV
	rjmp	update
	
invert:
	;perform last requested operation (if any), then calculate reciprocal value of result
	rcall	math_inv
	rjmp	update
	
sqrt:
	;perform last requested operation (if any), then calculate sqare root of result
	rcall	math_sqrt
	rjmp	update
	
ce:
	;clear input
	rcall	clear_input
	ori	flags, 1<<AnyInput | 1<<ManualInput
	rcall	clear_io
	rjmp	update
	
memclr:
	;clear memory
	rcall	clear_mem
	rcall	io_updatesymbols
	rjmp	waitkey
	
memrcl:
	;recall memory
	rcall	clear_input
	ori	flags, 1<<AnyInput
	rcall	copy_mem_to_io
	rjmp	update
	
memadd:
	;add result to memory
	sbrc	flags, AnyInput
	rcall	equal_subroutine
	rcall	copy_io_to_temp ;backup num_io
	rjmp	memaddsub
	
memsub:
	;subtract result from memory (by adding negative of result)
	sbrc	flags, AnyInput
	rcall	equal_subroutine
	rcall	copy_io_to_temp ;backup num_io
	ldi	temp, 1<<num_io_neg ;flip sign of num_io
	eor	flags, temp
	;(fall though to memaddsub)
	
memaddsub:
	;add num_io to num_op1, copy result to memory (and check if zero)
	;and restore last result from num_temp
	rcall	copy_op1_to_start ;backup num_op1
	rcall	swap_io_op1
	rcall	copy_mem_to_io
	rcall	math_add
	rcall	copy_io_to_mem
	rcall	copy_start_to_op1 ;restore num_op1
	rcall	copy_temp_to_io ;restore num_io
	;set flag if memory is now zero ("empty")
	ori	flags2, 1<<MemZero
	ldi	ZL, mem
	ldi	count, DISPSIZE
memaddsub_zero:
	ld	temp, Z+
	cpi	temp, 0
	breq	PC+2
	andi	flags2, ~(1<<MemZero)
	dec	count
	brne	memaddsub_zero
	rjmp	update
	
percent:
	sbrc	flags2, OP_EQUAL
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
	andi	flags, ~(1<<ShiftRounded) ;divide percentage by 100
	ldi	ZL, num_io
	rcall	math_shiftright_round
	ldi	ZL, num_io
	rcall	math_shiftright_round
	sbrc	flags2, OP_ADD
	rjmp	percent_add
	sbrs	flags2, OP_SUB
	rjmp	equal
	;subtraction: swap percentage's sign, then add
	ldi	temp, 1<<num_io_neg
	eor	flags, temp
percent_add:
	;addition
	rcall	copy_op1_to_temp
	rcall	clear_op1
	ldi	ZL, num_op1+(DISPSIZE/2-1)
	ldi	temp, 0x01
	st	Z, temp
	rcall	math_add ;add percentage to 1 (subtract from 1 if percentage is negative)
	rcall	copy_temp_to_op1
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
	ori	flags2, 1<<OP_MUL
	;(fall through to equal)
	
equal:
	ori	flags, 1<<AnyInput ;last result can be used as 2nd operand too
	rcall	equal_subroutine
	rjmp	update
	
equal_subroutine:
	;perform last requested operation, keep operation flag and add "equal" flag
	cpi	flags2, 1<<OP_ADD
	brlo	equal_perform ;no operation => call perform_operation anyway (clear input)
	sbrc	flags2, OP_EQUAL ;calculation w/ constant from op1 => swap operands
	rcall	swap_io_op1
equal_perform:
	rcall	copy_io_to_start ;backup num_io (2nd operand)
	rcall	perform_operation
	rcall	copy_start_to_op1 ;store 2nd operand as constant
	ori	flags2, 1<<OP_EQUAL
	ret
	
digit:
	;entered a digit
	cpi	inplen, DISPSIZE ;display is already full, can't enter any more
	brlo	PC+2
	rjmp	waitkey
	sbrc	flags, ManualInput ;first digit => clear num_io
	rjmp	digit_process
	rcall	clear_io
	rcall	clear_input
digit_process:
	ori	flags, 1<<ManualInput | 1<<AnyInput
	cpi	temp, 0
	breq	PC+2
	ori	flags, 1<<InputNonzero
	sbrc	flags, InputNonzero ;don't add leading zeros to number length
	inc	inplen
	;append digit to existing number
	cpi	decimals, 1
	brsh	digit_decimal
	;integer digit: shift num_io left one place and put new digit as units
	ldi	ZL, num_io
	push	temp
	rcall	math_shiftleft
	pop	temp
	ldi	ZL, num_io+(DISPSIZE/2-1)
	ld	temp2, Z
	andi	temp2, 0xF0 ;TODO: neccessary?
	or	temp2, temp
	st	Z, temp2
	rjmp	digit_end
digit_decimal:
	;fractional digit: do not shift, append digit after last digit
	ldi	ZL, num_io+(DISPSIZE/2-1) ;pointer to units digit
	mov	temp2, decimals
	inc	temp2
	lsr	temp2
	add	ZL, temp2
	ld	temp2, Z
	sbrc	decimals, 0
	swap	temp2
	andi	temp2, 0xF0 ;TODO: neccessary?
	or	temp2, temp
	sbrc	decimals, 0
	swap	temp2
	st	Z, temp2
	inc	decimals
digit_end:
	;calculate first displayed digit & length and update display
	ldi	ZL, num_io
	rcall	numfirst
	cpi	first, DISPSIZE ;display at least 1 digit before decimal point
	brlo	PC+2
	ldi	first, DISPSIZE-1
	mov	length, inplen
	cpi	inplen, 0
	brne	PC+2
	inc	length
	rcall	io_update
	rjmp	waitkey
	
;===============================================================================

perform_operation:
	;perform requested operation with num_op1 and num_io
	sbrs	flags, AnyInput ;do nothing if second operand is missing
	ret
	sbrc	flags2, OP_ADD
	rcall	math_add
	sbrc	flags2, OP_SUB
	rcall	math_sub
	sbrc	flags2, OP_MUL
	rcall	math_mul
	sbrc	flags2, OP_DIV
	rcall	math_div

clear_input:
	;clear number input helper registers (-> enter new number)
	andi	flags, ~(1<<ManualInput | 1<<AnyInput | 1<<InputNonzero)
	ldi	decimals, 0
	ldi	inplen, 0
	ret
	
;--------------------

clear_io:
	;clear num_io
	andi	flags, ~(1<<num_io_neg)
	ldi	ZL, num_io
	rjmp	clearnum

clear_op1:
	;clear num_op1
	andi	flags, ~(1<<num_op1_neg)
	ldi	ZL, num_op1
	rjmp	clearnum

clear_temp:
	;clear num_temp
	andi	flags, ~(1<<num_temp_neg)
	ldi	ZL, num_temp
	rjmp	clearnum
	
clear_mem:
	;clear memory
	andi	flags2, ~(1<<MemNeg)
	ori	flags2, 1<<MemZero
	ldi	ZL, mem
	;(fall through to clearnum)
	
clearnum:
	ldi	count, DISPSIZE
	ldi	temp2, 0
clearnum_loop:
	st	Z+, temp2
	dec	count
	brne	clearnum_loop
	ret
	
;--------------------

swap_io_op1:
	ldi	ZL, num_io
	ldi	count, DISPSIZE
swap_io_op1_loop:
	ld	temp, Z
	ldd	temp2, Z+(num_op1-num_io)
	std	Z+(num_op1-num_io), temp
	st	Z+, temp2
	dec	count
	brne	swap_io_op1_loop
	mov	temp, flags
	andi	flags, ~(1<<num_io_neg | 1<<num_op1_neg)
	sbrc	temp, num_op1_neg
	ori	flags, 1<<num_io_neg
	sbrc	temp, num_io_neg
	ori	flags, 1<<num_op1_neg
	ret

;--------------------

copy_io_to_op1:
	;copy num_op1 to num_io (with sign)
	andi	flags, ~(1<<num_op1_neg)
	sbrc	flags, num_io_neg
	ori	flags, 1<<num_op1_neg
	ldi	YL, num_io
	ldi	ZL, num_op1
	rjmp	copynum

;--------------------

copy_io_to_temp:
	;copy num_io to num_temp (with sign)
	andi	flags, ~(1<<num_temp_neg)
	sbrc	flags, num_io_neg
	ori	flags, 1<<num_temp_neg
	ldi	YL, num_io
	ldi	ZL, num_temp
	rjmp	copynum

;--------------------

copy_io_to_start:
	;copy num_op1 to iteration value (with sign)
	andi	flags2, ~(1<<num_start_neg)
	sbrc	flags, num_io_neg
	ori	flags2, 1<<num_start_neg
	ldi	YL, num_io
	ldi	ZL, num_start
	rjmp	copynum

;--------------------

copy_io_to_iter:
	;copy num_op1 to iteration value (no sign!)
	ldi	YL, num_io
	ldi	ZL, num_iter
	rjmp	copynum

;--------------------

copy_io_to_mem:
	;copy num_io to memory (with sign)
	andi	flags2, ~(1<<MemNeg)
	sbrc	flags, num_io_neg
	ori	flags2, 1<<MemNeg
	ldi	YL, num_io
	ldi	ZL, mem
	rjmp	copynum

;--------------------

copy_op1_to_temp:
	;copy num_op1 to num_temp (with sign)
	andi	flags, ~(1<<num_temp_neg)
	sbrc	flags, num_op1_neg
	ori	flags, 1<<num_temp_neg
	ldi	YL, num_op1
	ldi	ZL, num_temp
	rjmp	copynum

;--------------------

copy_op1_to_start:
	;copy num_io to num_start (no sign)
	andi	flags2, ~(1<<num_start_neg)
	sbrc	flags, num_op1_neg
	ori	flags2, 1<<num_start_neg
	ldi	YL, num_op1
	ldi	ZL, num_start
	rjmp	copynum

;--------------------

copy_temp_to_io:
	;copy num_temp to num_io (with sign)
	andi	flags, ~(1<<num_io_neg)
	sbrc	flags, num_temp_neg
	ori	flags, 1<<num_io_neg
	ldi	YL, num_temp
	ldi	ZL, num_io
	rjmp	copynum

;--------------------

copy_temp_to_op1:
	;copy num_temp to num_op1 (with sign)
	andi	flags, ~(1<<num_op1_neg)
	sbrc	flags, num_temp_neg
	ori	flags, 1<<num_op1_neg
	ldi	YL, num_temp
	ldi	ZL, num_op1
	rjmp	copynum

;--------------------

copy_start_to_op1:
	;copy num_start to num_op1 (with sign)
	andi	flags, ~(1<<num_op1_neg)
	sbrc	flags2, num_start_neg
	ori	flags, 1<<num_op1_neg
	ldi	YL, num_start
	ldi	ZL, num_op1
	rjmp	copynum

;--------------------

copy_iter_to_op1:
	;copy iteration value to num_io (no sign!)
	ldi	YL, num_iter
	ldi	ZL, num_op1
	rjmp	copynum

;--------------------

copy_mem_to_io:
	;copy memory to num_io (with sign)
	andi	flags, ~(1<<num_io_neg)
	sbrc	flags2, MemNeg
	ori	flags, 1<<num_io_neg
	ldi	YL, mem
	ldi	ZL, num_io
	rjmp	copynum

;--------------------

copynum:
	ldi	count, DISPSIZE
copynum_loop:
	ld	temp2, Y+
	st	Z+, temp2
	dec	count
	brne	copynum_loop
	ret
	