;===============================================================================
;  input/output module: 7-segment LED or VFD (up to 15 digits) and 4x6 keypad
;===============================================================================
;
;Ports:
;  PB0-7 = Segments a..g, dp (active low for LED)
;  PD0-3 = Keypad Columns 1..4 (external pulldowns for VFD)
;  PD4-6 = LED Digit/Keypad Row Selection 0..2
;  PA0   = LED Digit/Keypad Row Selection 3
;
;Digit Selector (LED):                                 _______
;                         (<= 8 digits: 74HC138, PA0 = Digit 0 = ENABLE[Pin6])
;                     4   ,------------------,          (>= 9 digits: 74HC154)
;  PA0 PD6 PD5 PD4 ===/===| Binary to Hex    |
;                         '-oooooooooooooooo-'
;                    _____  ||||||||||||||||
;                    Digit 0 .............. 15  <--  active low outputs
;
;Digit Selector (VFD):
;
;                     4   ,------------------,
;  PA0 PD6 PD5 PD4 ===/===| Binary to Hex    |
;                         '------------------'
;                           ||||||||||||||||
;                    Digit 0 .............. 15  <--  active high outputs
;
;LED Display:
;  Example for 8 digits: VQE12, 3x VQE14, VQE12-upside-down
;
;     0 15-(n-1) ... 14  15      <-- Digit 0, 15-(n-1)...15 (n = DISPSIZE)
;     |   |          |   |
;     R   R   ....   R   R       <-- Resistors (~ 1k)
;    /   /          /   /
;   / ,-/-o----//--/-o-/-o-------(+5V)
;  /BE|/BE|       /BE|/BE|
;  `|< `|<    ... `|< `|<        <-- PNP Transistors (e.g. BC327-40)
;    C|  C|         C|  C|
;     o---)----------)---)---,
;     |   |          |   |   |
;   ,---,---,--//--,---,---,---,   8
;   | - | 8.| .... | 8.| 8 |^+ |===/===[R]=== PB0..7, R = 8x ~33 Ohms
;   '---'---'--//--'---'---'---'
;
;Vacuum Fluorescent Display:
;
;     0 15-(n-1) ... 14  15      <-- Digit 0, 15-(n-1)...15 (n = DISPSIZE)
;     |   |          |   |
;  ,--'---'------//--'---'--,
;  | VFD DRIVER WITH        |----- (+24V)
;  | ACTIVE HIGH INPUTS     |
;  '--,---,------//--,---,--'
;     |   |          |   |                      (+24V)
;     |   |          |   |                        |
;   ,---,---,--//--,---,---,---,   8   ,----------'--,
;   |ME-| 8.| .... | 8.| 8 |^+ |===/===| VFD DDRIVER |=== PB0..7
;   '---'---'-//---'---'---'---'       '-------------'
;
;   leftmost/rightmost displays:
;       Operation Add, Sub (minus sign)                 : Segment a
;       Operation Add, Mul, Div (plus sign upper stroke): Segment b
;       Operation Add, Div (plus sign lower stroke)     : Segment c
;       Error                                           : Segment f
;       Result is negative (left minus sign)            : Segment g
;       Memory =/= 0                                    : Segment dp
;
;Keypad (the layout can be changed below, search for "keytable:"):
;  ,---,---,---,---,
;  |1/x|sqr|   |   |---|<|--- Digit 9
;  +---+---+---+---+
;  |MR |MC |CE |AC |---|<|--- Digit 10
;  +---+---+---+---+
;  |M+ |M- |+/-| / |---|<|--- Digit 11
;  +---+---+---+---+
;  | 7 | 8 | 9 | * |---|<|--- Digit 12
;  +---+---+---+---+
;  | 4 | 5 | 6 | - |---|<|--- Digit 13
;  +---+---+---+---+
;  | 1 | 2 | 3 | + |---|<|--- Digit 14
;  +---+---+---+---+
;  | 0 | . | % | = |---|<|--- Digit 15
;  '---'---'---'---'
;    |   |   |   |     Reverse Diodes for LED
;   PD0 PD1 PD2 PD3
;    |   |   |   |
;   _|_ _|_ _|_ _|_
;   | | | | | | | | 4x 10..50k
;   |_| |_| |_| |_| Pulldowns (omit for LED)
;    |   |   |   |
;    '---o---o---o----- (GND)

;number of display digits:
;  all functions available: 7 <= DISPSIZE <= 15
;  no sqrt and 1/x:         6 <= DISPSIZE <= 15
.equ	DISPSIZE = 8
;display type:
;  LED displays: VFD = 0
;  VFD tube(s):  VFD = 1
.equ	VFD = 0

.def	seldigit = R26 ;XL
.def	key = R27 ;XH

;===============================================================================
; interrupts
;===============================================================================

.org OVF0addr
	;timer 0 overflow interrupt
	push	temp
	push	temp2
	push	ZL
	push	ZH
	in	temp, SREG
	push	temp
	
	;read keys in current row
	cpi	seldigit, 9
	brlo	io_ovf0_keyend
	ldi	ZL, LOW(keytable*2) ;set pointer to beginning of current row
	ldi	ZH, HIGH(keytable*2)
	mov	temp2, seldigit
	subi	temp2, 9
	lsl	temp2
	lsl	temp2
	add	ZL, temp2
	ldi	temp2, 0
	adc	ZH, temp2
	
	;check if a key in the current row is pressed
	in	temp, PIND
	ldi	temp2, 0xFF
	.if (VFD == 0)
	com	temp
	.endif
	sbrc	temp, 0
	ldi	temp2, 0
	sbrc	temp, 1
	ldi	temp2, 1
	sbrc	temp, 2
	ldi	temp2, 2
	sbrc	temp, 3
	ldi	temp2, 3
	cpi	temp2, 0xFF ;no key in current row pressed if still 0xFF
	breq	io_ovf0_keyrelease
	;load keycode of pressed key from layout table
	add	ZL, temp2
	ldi	temp2, 0
	add	ZH, temp2
	lpm	key, Z
	rjmp	io_ovf0_keyend
	
io_ovf0_keyrelease:
	;check if a key in the current row was released
	ldi	temp2, 4
io_ovf0_keyrelease_loop:
	;read key codes in current row and compare to last keypress
	lpm	temp, Z+
	cp	key, temp
	breq	io_ovf0_keyreset
	dec	temp2
	brne	io_ovf0_keyrelease_loop
	rjmp	io_ovf0_keyend
io_ovf0_keyreset:
	ldi	key, 0xFF
io_ovf0_keyend:

	;increment digit pointer
	inc	seldigit
	andi	seldigit, 0x0F
	cpi	seldigit, 0
	breq	io_ovf0_incend
	cpi	seldigit, 15-(DISPSIZE-1)
	brsh	io_ovf0_incend
	ldi	seldigit, 15-(DISPSIZE-1)
io_ovf0_incend:

	;display off
	.if (VFD == 0)
	ldi	temp, 0xFF
	.else
	ldi	temp, 0x00
	.endif
	out	PORTB, temp
	rcall	shortwait
	
	;select new digit
	mov	temp, seldigit
	swap	temp
	.if (VFD == 0)
	ori	temp, 0x0F
	.endif
	ldi	temp2, 0x00
	sbrc	seldigit, 3
	ldi	temp2, 0x01
	out	PORTD, temp
	out	PORTA, temp2
	rcall	shortwait
	
	;output new digit
	ld	temp, X ;XL = seldigit
	out	PORTB, temp
	
	;return to main program
	pop	temp
	out	SREG, temp
	pop	ZH
	pop	ZL
	pop	temp2
	pop	temp
	reti
	
;--------------------

shortwait:
	ldi	temp, 20
shortwait_loop:
	dec	temp
	brne	shortwait_loop
	ret

;--------------------
	
;keypad layout table, 4 cols x 7 rows (0xff = no function)
;this table can be changed if you with to use a different layout
;keytable: ;original layout
;	;   PD0         PD1         PD2          PD3
;	.db KEY_INV,    KEY_SQRT,   0xff,        0xff      ;--|<|-- digit 9
;	.db KEY_MEMRCL, KEY_MEMCLR, KEY_CE,      KEY_AC    ;--|<|-- digit 10
;	.db KEY_MEMADD, KEY_MEMSUB, KEY_NEG,     KEY_DIV   ;--|<|-- digit 11
;	.db 7,          8,          9,           KEY_MUL   ;--|<|-- digit 12
;	.db 4,          5,          6,           KEY_SUB   ;--|<|-- digit 13
;	.db 1,          2,          3,           KEY_ADD   ;--|<|-- digit 14
;	.db 0,          KEY_DOT,    KEY_PERCENT, KEY_EQUAL ;--|<|-- digit 15

keytable: ;alternative layout
	;   PD0         PD1         PD2          PD3
	.db KEY_INV,    KEY_SQRT,   0xff,        0xff      ;--|<|-- digit 9
	.db KEY_MEMRCL, KEY_MEMCLR, KEY_CE,      KEY_AC    ;--|<|-- digit 10
	.db KEY_MEMADD, KEY_MEMSUB, KEY_PERCENT, KEY_NEG   ;--|<|-- digit 11
	.db 7,          8,          9,           KEY_DIV   ;--|<|-- digit 12
	.db 4,          5,          6,           KEY_MUL   ;--|<|-- digit 13
	.db 1,          2,          3,           KEY_SUB   ;--|<|-- digit 14
	.db 0,          KEY_DOT,    KEY_EQUAL,   KEY_ADD   ;--|<|-- digit 15

;(diodes drawn for VFD, reverse for LED)

;===============================================================================
; public functions (called from the main program)
;===============================================================================

io_init:
	;initialize I/O interface (called after power-on)
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x00
	out	PORTA, temp
	ldi	temp, 0xFF
	out	DDRB, temp
	.if (VFD == 0)
	ldi	temp, 0xFF
	.else
	ldi	temp, 0x00
	.endif
	out	PORTB, temp
	ldi	temp, 0x70
	out	DDRD, temp
	.if (VFD == 0)
	ldi	temp, 0x0F
	.else
	ldi	temp, 0x00
	.endif
	out	PORTD, temp
	
	ldi	temp, 0x02 ;Clk/8
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	ldi	seldigit, 15-(DISPSIZE-1)
	ldi	key, 0xFF
	sei
	ret
	
;--------------------

io_update:
	;update display (called after number or operation changed)
	sbrs	flags, Error ;display "Error" if an error occured
	rjmp	io_update_noerror
	ldi	ZL, 15-(DISPSIZE-1)
	.if (VFD == 0)
	ldi	temp, 0x86 ;'E'
	st	Z+, temp
	ldi	temp2, 0xAF ;'r'
	st	Z+, temp2
	st	Z+, temp2
	ldi	temp, 0xA3 ;'o'
	st	Z+, temp
	st	Z+, temp2 ;'r'
	ldi	count, DISPSIZE-5
	ldi	temp, 0xFF ;all segments off
	.else
	ldi	temp, 0x79 ;'E'
	st	Z+, temp
	ldi	temp2, 0x50 ;'r'
	st	Z+, temp2
	st	Z+, temp2
	ldi	temp, 0x5C ;'o'
	st	Z+, temp
	st	Z+, temp2 ;'r'
	ldi	temp, 0x00 ;all segments off
	.endif
io_update_error:
	st	Z+, temp
	cpi	ZL, 16
	brne	io_update_error
	.if (VFD == 0)
	ldi	temp, ~0x20 ;display Error symbol
	.else
	ldi	temp, 0x20 ;display Error symbol
	.endif
	mov	R0, temp
	ret
io_update_noerror:
	;output spaces to right-align numbers
	push	length
	mov	ZL, first
	lsr	ZL
	subi	ZL, -num_io
	ldi	YL, 15-(DISPSIZE-1)
	.if (VFD == 0)
	ldi	temp, 0xFF ;all segments off
	.else
	ldi	temp, 0x00 ;all segments off
	.endif
	ldi	count, DISPSIZE
	sub	count, length
io_update_space:
	cpi	count, 0
	breq	io_update_num
	cpi	count, 1 ;use last digit before number (if any) for sign
	brne	io_update_space_nosign
	sbrc	flags, num_io_neg
	.if (VFD == 0)
	ldi	temp, ~0x40 ;minus sign
	.else
	ldi	temp, 0x40 ;minus sign
	.endif
io_update_space_nosign:
	st	Y+, temp
	dec	count
	rjmp	io_update_space
io_update_num:
	;output number digits (and decimal point after the units digit)
	ld	temp, Z
	sbrc	first, 0
	inc	ZL
	sbrs	first, 0
	swap	temp
	andi	temp, 0x0F
	push	ZL
	ldi	ZL, LOW(io_update_segments*2)
	ldi	ZH, HIGH(io_update_segments*2)
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z
	pop	ZL
	cpi	first, DISPSIZE-1
	brne	PC+2
	.if (VFD == 0)
	andi	temp, ~0x80 ;dp enabled
	.else
	ori	temp, 0x80 ;dp enabled
	.endif
	st	Y+, temp
	inc	first
	dec	length
	brne	io_update_num
	pop	length
	
io_updatesymbols:
	;output next operation, sign, and memory state (zero/nonzero)
	.if (VFD == 0)
	ldi	temp, 0xFF ;all segments off
;	sbrc	flags, AnyInput
;	rjmp	io_update_sign
	sbrc	flags2, OP_EQUAL
	rjmp	io_update_sign
	sbrc	flags2, OP_ADD
	ldi	temp, ~0x07 ;display minus and both strokes of plus sign
	sbrc	flags2, OP_SUB
	ldi	temp, ~0x01 ;display only minus sign
	sbrc	flags2, OP_MUL
	ldi	temp, ~0x02 ;display only upper stroke of plus sign ("dot")
	sbrc	flags2, OP_DIV
	ldi	temp, ~0x06 ;display both strokes of plus sign ("colon")
	.else
	ldi	temp, 0x00 ;all segments off
	sbrc	flags, AnyInput
	rjmp	io_update_sign
	sbrc	flags2, OP_EQUAL
	rjmp	io_update_sign
	sbrc	flags2, OP_ADD
	ldi	temp, 0x07 ;display minus and both strokes of plus sign
	sbrc	flags2, OP_SUB
	ldi	temp, 0x01 ;display only minus sign
	sbrc	flags2, OP_MUL
	ldi	temp, 0x02 ;display only upper stroke of plus sign ("dot")
	sbrc	flags2, OP_DIV
	ldi	temp, 0x06 ;display both strokes of plus sign ("colon")
	.endif
io_update_sign:
	cpi	length, DISPSIZE
	brlo	io_update_sign_end
	sbrc	flags, num_io_neg
	.if (VFD == 0)
	andi	temp, ~0x40 ;display left-hand minus sign (negative number)
	.else
	ori	temp, 0x40 ;display left-hand minus sign (negative number)
	.endif
io_update_sign_end:
	sbrs	flags2, MemZero
	.if (VFD == 0)
	andi	temp, ~0x80 ;display 'M' symbol
	.else
	ori	temp, 0x80 ;display 'M' symbol
	.endif
	mov	R0, temp
	
	ret
	
io_update_segments:
	.if (VFD == 0)
	;      0      1      2      3      4      5      6      7
	.db ~0x3f, ~0x06, ~0x5b, ~0x4f, ~0x66, ~0x6d, ~0x7d, ~0x07
	;      8      9
	.db ~0x7f, ~0x6f
	.else
	;     0     1     2     3     4     5     6     7
	.db 0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07
	;     8     9
	.db 0x7f, 0x6f
	.endif
	
;--------------------

io_getkey:
	;wait for keypress and return keycode
	cpi	key, 0xFF ;wait for release
	brne	io_getkey
io_getkey_keypress: ;wait for keypress
	cpi	key, 0xFF
	breq	io_getkey_keypress
	mov	temp, key
	ret
