;===============================================================================
;  math functions
;===============================================================================

numfirst:
	;return position of first nonzero digit of the number starting at ZL
	;(if none, return DISPSIZE*2 and set Z flag)
	ldi	first, 0
	ldi	count, DISPSIZE
numfirst_loop:
	ld	temp, Z+
	cpi	temp, 0x10
	brsh	numfirst_end ;Z flag set or cleared!
	inc	first
	cpi	temp, 0
	brne	numfirst_end ;Z flag cleared
	inc	first
	dec	count
	brne	numfirst_loop
	;Z flag set
	ret
numfirst_end:
	clz ;clear Z flag
	ret
	
;--------------------

math_compare:
	;compare num_op1 to num_io and set Z (equal) and C (num_op1 < num_io) flags
	ldi	ZL, num_io
	ldi	count, DISPSIZE
math_compare_loop:
	ldd	temp, Z+(num_op1-num_io)
	ld	temp2, Z+
	cp	temp, temp2
	brne	math_compare_end ;num_op1 != num_io, flags already set by 'cp'
	dec	count
	brne	math_compare_loop
	;num_op1 == num_io (flags already set [Z=1 by last 'dec', C=0 by 'cp'])
math_compare_end:
	ret

;--------------------

math_shiftright_round:
	;shift right (divide by 10) the number starting at ZL
	subi	ZL, -(DISPSIZE-1) ;pointer to last digit
	mov	ZH, ZL
	ldi	count, DISPSIZE
	ld	ZH, Z ;keep shifted-out digit for rounding
	andi	ZH, 0x0F
math_shiftright_loop:
	ld	temp, Z
	swap	temp
	ld	temp2, -Z
	swap	temp2
	andi	temp, 0x0F
	andi	temp2, 0xF0
	or	temp, temp2
	std	Z+1, temp
	dec	count
	brne	math_shiftright_loop
	andi	temp, 0x0F ;clear first digit
	std	Z+1, temp
	
	;rounding
	ldi	temp, 5 ;round up if shifted-out digit >= 5
	sbrc	flags, ShiftRounded
	ldi	temp, 6 ;previously rounded up: shifted-out digit must be >= 6
	andi	flags, ~(1<<ShiftRounded)
	cp	ZH, temp
	brlo	math_shiftright_end
	;round up: add 1 to last digit
	subi	ZL, -(DISPSIZE+1)
math_shiftright_inc:
	ld	temp2, -Z
	subi	temp2, -0x07
	brhc	math_shiftright_inc_hc
	subi	temp2, 0x06
	rjmp	math_shiftright_inc_end
math_shiftright_inc_hc: ;halfcarry set: upper nibble incremented
	subi	temp2, -0x60
	brcc	math_shiftright_inc_c
	subi	temp2, 0x60
	rjmp	math_shiftright_inc_end
math_shiftright_inc_c: ;carry set: increment lower nibble of previous byte
	st	Z, temp2
	rjmp	math_shiftright_inc ;loop (overflow impossible: first digit is 0)
math_shiftright_inc_end:
	st	Z, temp2
	ori	flags, 1<<ShiftRounded
math_shiftright_end:
	ret
	
;--------------------

math_shiftleft:
	;shift left (multiply by 10) the number starting at ZL
	ld	temp2, Z
	cpi	temp2, 0x10
	brsh	math_overflow
	ldi	count, DISPSIZE
math_shiftleft_loop:
	ld	temp, Z
	ldd	temp2, Z+1
	andi	temp, 0x0F
	andi	temp2, 0xF0
	swap	temp
	swap	temp2
	or	temp, temp2
	st	Z+, temp
	dec	count
	brne	math_shiftleft_loop
	andi	temp, 0xF0 ;clear last digit
	st	-Z, temp
	ret
	
;--------------------

math_sub:
	;subtraction: num_op1 = num_op1 - num_io, done by adding num_op1 and -num_io
	ldi	temp, 1<<num_io_neg
	eor	flags, temp ;flip sign of num_io
	;(fall through to math_add)

;--------------------

math_add:
	;addition: num_io = num_op1 + num_io
	;handle negative numbers this way:
	;  same sign:       num_io = (num_io's sign) (|num_op1| + |num_io|)
	;  different signs: num_io = (num_op1's sign) (|num_op1| - |num_io|)
	ldi	temp, 0
	sbrc	flags, num_op1_neg
	inc	temp
	sbrc	flags, num_io_neg
	inc	temp
	cpi	temp, 1
	brne	math_add_abs
	;subtract absolute values, copy result from num_op1 to num_io
	rcall	math_sub_abs
	rcall	swap_io_op1 ;num_op1 is expected to be destroyed, so value doesn't matter
	ret
	
math_add_abs:
	;add absolute values: num_io = (num_io's sign) (|num_op1| + |num_io|)
	clt ;T flag is used as decimal Carry Bit
	ldi	ZL, num_io+DISPSIZE
	ldi	count, DISPSIZE
math_add_loop:
	ld	temp2, -Z
	ldd	temp, Z+(num_op1-num_io)
	;packed BCD addition
	brtc	PC+2
	inc	temp2
	clt
	add	temp, temp2
	brcc	PC+2
	set
	brhs	math_add_1
	subi	temp, -6
	brhc	math_add_2
	subi	temp, 12
math_add_1:
	subi	temp, -6
math_add_2:
	subi	temp, -0x60
	brcc	math_add_3
	brts	math_add_end
	subi	temp, 0x60
	rjmp	math_add_end
math_add_3:
	set
math_add_end:
	st	Z, temp
	dec	count
	brne	math_add_loop
	brts	math_overflow ;carry bit set => overflow
	ret
	
;--------------------

math_overflow:
	ori	flags, 1<<Error
	ret
	
;--------------------
	
math_sub_abs:
	;subtract absolute values: num_op1 = (num_op1's sign) (|num_op1| - |num_io|)
	;NOTE: result goes to num_op1, NOT num_io
	rcall	math_compare
	brsh	math_sub_start
	;|num_op1| < |num_io|: swap and flip sign of result
	push	flags
	rcall	swap_io_op1
	pop	temp
	andi	flags, ~(1<<num_op1_neg)
	sbrc	temp, num_op1_neg
	ori	flags, 1<<num_op1_neg
	ldi	temp, 1<<num_op1_neg
	eor	flags, temp
	
math_sub_start:
	;subtract |num_io| from |num_op1| (at this point, |num_op1| >= |num_io|)
	clt ;T flag is used as decimal Carry bit
	ldi	ZL, num_io+DISPSIZE
math_sub_loop:
	ld	temp2, -Z
	ldd	temp, Z+(num_op1-num_io)
	;packed BCD subtraction
	brtc	PC+2
	inc	temp2
	clt
	sub	temp, temp2
	brcc	PC+2
	set
	brhc	PC+2
	subi	temp, 0x06
	brtc	PC+2
	subi	temp, 0x60
	std	Z+(num_op1-num_io), temp
	cpi	ZL, num_io
	brne	math_sub_loop
	ret
	
;--------------------

math_mul:
	;multiply num_op1 and num_io
	rcall	copy_io_to_temp ;copy second factor to num_temp and clear num_io
	rcall	clear_io
	
	ldi	temp, 1<<num_io_neg ;result's sign = (factor1's) XOR (factor2's)
	sbrc	flags, num_temp_neg
	or	flags, temp
	sbrc	flags, num_op1_neg
	eor	flags, temp
	
	;find first nonzero digit in first factor (num_op1)
	ldi	ZL, num_op1
	rcall	numfirst
	brne	PC+2
	ret ;first factor is 0 => result is 0
	push	first
	
	;shift first factor so second digit is first nonzero digit
	andi	flags, ~(1<<ShiftRounded)
	cpi	first, 0
	brne	math_mul_shiftleft
	ldi	ZL, num_op1
	rcall	math_shiftright_round
	ldi	first, 1
	rjmp	math_mul_shiftend
math_mul_shiftleft:
	cpi	first, 1
	breq	math_mul_shiftend
	ldi	ZL, num_op1
	rcall	math_shiftleft
	dec	first
	rjmp	math_mul_shiftleft
math_mul_shiftend:
	
	;find first nonzero digit in second factor (num_temp)
	ldi	ZL, num_temp
	rcall	numfirst
	brne	math_mul_foundfirst
	;second factor is 0 => result is 0
	pop	temp ;remove data from stack
	ret
math_mul_foundfirst:
	push	first

	;add shifted 1st factor to result (1st digit of 2nd factor)-times,
	;then shift 1st factor right 1 place and continue with next digit
;	andi	flags, ~(1<<ShiftRounded)
	ldi	YL, num_temp
	mov	temp, first
	lsr	temp
	add	YL, temp
math_mul_add:
	ld	YH, Y
	sbrc	first, 0 ;odd digit -> next digit is in the next byte
	inc	YL
	sbrs	first, 0 ;odd digit -> right nibble
	swap	YH ;even digit -> left nibble
	andi	YH, 0x0F
	breq	math_mul_add_next
math_mul_add_loop:
	rcall	math_add_abs
	dec	YH
	brne	math_mul_add_loop
math_mul_add_next:
	ldi	ZL, num_op1
	rcall	math_shiftright_round
	inc	first
	cpi	first, DISPSIZE*2
	brlo	math_mul_add
	
	;shift result to correct position
	ldi	YH, DISPSIZE
	pop	temp
	sub	YH, temp
	pop	temp
	sub	YH, temp
	rjmp	math_muldiv_shiftresult
	
;--------------------

math_div:
	;divide num_op1 by num_io
	rcall	clear_temp ;clear num_temp for result
	ldi	temp, 0 ;result's sign = (dividend's) XOR (divisor's)
	sbrc	flags, num_op1_neg
	inc	temp
	sbrc	flags, num_io_neg
	inc	temp
	sbrc	temp, 0
	ori	flags, 1<<num_temp_neg
	
	;find first nonzero digit in divisor (num_io)
	ldi	ZL, num_io
	rcall	numfirst
	brne	math_div_defined
	ori	flags, 1<<Error ;divisor is 0 => result is undefined
	ret
math_div_defined:
	push	first
	
	;shift divisor to the very left
math_div_shift1:
	cpi	first, 0
	breq	math_div_shift1end
math_div_shift1_loop:
	ldi	ZL, num_io
	rcall	math_shiftleft
	dec	first
	brne	math_div_shift1_loop
math_div_shift1end:
	
	;find first nonzero digit in dividend (num_op1)
	ldi	ZL, num_op1
	rcall	numfirst
	brne	math_div_foundfirst
	;dividend is 0 => result is 0
	rcall	clear_io
	pop	temp ;remove data from stack
	ret
math_div_foundfirst:
	push	first
	
	;shift dividend to the very left
math_div_shift2:
	cpi	first, 0
	breq	math_div_shift2end
math_div_shift2_loop:
	ldi	ZL, num_op1
	rcall	math_shiftleft
	dec	first
	brne	math_div_shift2_loop
math_div_shift2end:

	;subtract while dividend > divisor and increment result digit,
	;then shift divisor right and continue with next result digit
	andi	flags, ~(1<<ShiftRounded)
	ldi	YL, num_temp
	ldi	YH, 0x10
math_div_loop:
	rcall	math_compare
	brlo	math_div_next ;dividend < divisor
math_div_subtract: ;subtract divisor from dividend, increment result
	rcall	math_sub_abs ;num_op1 - num_io -> num_op1
	ld	temp, Y
	add	temp, YH
	st	Y, temp
	rjmp	math_div_loop
math_div_next: ;continue with next digit
	ldi	ZL, num_io
	rcall	math_shiftright_round
	swap	YH ;toggle between adding 0x10 and 0x01
	sbrc	YH, 4 ;next byte if adding 0x10 comes next
	inc	YL
	cpi	YL, num_temp+DISPSIZE
	brlo	math_div_loop
	
	;copy result to num_io and shift to correct position
	rcall	copy_temp_to_io
	andi	flags, ~(1<<ShiftRounded)
	ldi	YH, -(DISPSIZE-1)
	pop	temp
	sub	YH, temp
	pop	temp
	add	YH, temp
	;(fall through to math_muldiv_shiftresult)
	
;--------------------
	
math_muldiv_shiftresult:
	;shift result left YH times (YH < 0: shift right)
	brmi	math_muldiv_resultright
	brne	math_muldiv_resultleft
	ret
math_muldiv_resultright:
	andi	flags, ~(1<<ShiftRounded)
math_muldiv_resultright_loop:
	ldi	ZL, num_io
	rcall	math_shiftright_round
	inc	YH
	brne	math_muldiv_resultright_loop
	ret
math_muldiv_resultleft:
	ldi	ZL, num_io
	rcall	math_shiftleft
	dec	YH
	brne	math_muldiv_resultleft
	ret

;--------------------

math_inv:
	;calculate reciprocal inverse of num_op1
	rcall	copy_op1_to_start ;backup op1
	rcall	clear_op1 ;load num_op1 with 1
	ldi	ZL, num_op1+(DISPSIZE/2-1)
	ldi	temp, 0x01
	st	Z, temp
	rcall	math_div ;calculate num_op1/num_io (1/x)
	rcall	copy_start_to_op1 ;restore op1
	ret

;--------------------

math_sqrt:
	;calculate square root of num_io  (overflow if num_io >= [999...9])
	;  initialization: iter(0) = start = num_io
	;  iteration: iter(i+1) = 0.5 * [iter(i) + start / iter(i)]
	;error if num_io is negative
	;TODO: backup and restore num_op1
	sbrc	flags, num_io_neg
	rjmp	math_overflow
	;return 0 if num_io is 0
	ldi	ZL, num_io
	rcall	numfirst
	breq	math_sqrt_end
	
	;save start value
	rcall	copy_io_to_op1
	rcall	copy_op1_to_start
	
	ldi	count, DISPSIZE*3 ;maximum number of iterations
math_sqrt_iteration:
	;iteration loop
	subi	count, 1
	brcs	math_sqrt_iteration_end
	push	count
	rcall	copy_io_to_iter
	rcall	copy_start_to_op1
	rcall	math_div ;a / x(i)
	rcall	copy_iter_to_op1
	rcall	math_add ;x(i) + ...
	rcall	clear_op1
	ldi	ZL, num_op1+(DISPSIZE/2)
	ldi	temp, 0x50
	st	Z, temp
	rcall	math_mul ;0.5 * ... (... / 2)
	rcall	copy_iter_to_op1
	rcall	math_compare ;iteration finished if x(i+1) = x(i)
	pop	count
	sbrs	flags, Error ;abort on error to prevent endless loop
	brne	math_sqrt_iteration
math_sqrt_iteration_end:
	;TODO: remove after backup/restore of op1 during iteration:
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV | 1<<OP_EQUAL)
math_sqrt_end:
	ret
	